__name__ = 'pysgems'
__author__ = 'Robin Thibaut'

from . import algo
from . import base
from . import dis
from . import examples
from . import io
from . import plot
from . import sgems
from . import utils
